"""
soundscapy.audio.metrics
========================

This module provides functions for calculating various acoustic and psychoacoustic metrics
for audio signals. It includes implementations for single-channel and two-channel signals,
as well as wrapper functions for different libraries such as python-acoustics, MoSQITo,
and scikit-maad.

Functions
---------
_stat_calcs : Calculate various statistics for a time series array.
mosqito_metric_1ch : Calculate a MoSQITo psychoacoustic metric for a single channel signal.
maad_metric_1ch : Run a metric from the scikit-maad library on a single channel signal.
pyacoustics_metric_1ch : Run a metric from the pyacoustics library on a single channel object.
pyacoustics_metric_2ch : Run a metric from the python acoustics library on a Binaural object.
mosqito_metric_2ch : Calculate metrics from MoSQITo for a two-channel signal.
maad_metric_2ch : Run a metric from the scikit-maad library on a binaural signal.
prep_multiindex_df : Prepare a MultiIndex dataframe from a dictionary of results.
add_results : Add results to a MultiIndex dataframe.
process_all_metrics : Process all metrics specified in the analysis settings for a binaural signal.

Notes
-----
This module relies on external libraries such as numpy, pandas, maad, mosqito, and scipy.
Ensure these dependencies are installed before using this module.
"""

import concurrent.futures
import multiprocessing as mp
from typing import Dict, List, Optional, Tuple, Union
from loguru import logger

import numpy as np
import pandas as pd
from maad.features import all_spectral_alpha_indices, all_temporal_alpha_indices
from maad.sound import spectrogram
from mosqito.sq_metrics import (
    loudness_zwtv,
    roughness_dw,
    sharpness_din_from_loudness,
    sharpness_din_perseg,
    sharpness_din_tv,
)
from scipy import stats

from .analysis_settings import AnalysisSettings

DEFAULT_LABELS = {
    "LZeq": "LZeq",
    "LCeq": "LCeq",
    "LAeq": "LAeq",
    "Leq": "Leq",
    "SEL": "SEL",
    "loudness_zwtv": "N",
    "roughness_dw": "R",
    "sharpness_din_perseg": "S",
    "sharpness_din_from_loudness": "S",
    "sharpness_din_tv": "S",
}


def _stat_calcs(
    label: str, ts_array: np.ndarray, res: dict, statistics: List[Union[int, str]]
) -> dict:
    """
    Calculate various statistics for a time series array and add them to a results dictionary.

    Parameters
    ----------
    label : str
        Base label for the statistic in the results dictionary.
    ts_array : np.ndarray
        1D numpy array of the time series data.
    res : dict
        Existing results dictionary to update with new statistics.
    statistics : List[Union[int, str]]
        List of statistics to calculate. Can include percentiles
        (as integers) and string identifiers for other statistics
        (e.g., "avg", "max", "min", "kurt", "skew").

    Returns
    -------
    dict
        Updated results dictionary with newly calculated statistics.

    Examples
    --------
    >>> # xdoctest: +REQUIRES(env:AUDIO_DEPS='1')
    >>> ts = np.array([1, 2, 3, 4, 5])
    >>> res = {}
    >>> updated_res = _stat_calcs("metric", ts, res, [50, "avg", "max"])
    >>> print(updated_res)
    {'metric_50': 3.0, 'metric_avg': 3.0, 'metric_max': 5}
    """
    logger.debug(f"Calculating statistics for {label}")
    for stat in statistics:
        try:
            if stat in ("avg", "mean"):
                res[f"{label}_{stat}"] = ts_array.mean()
            elif stat == "max":
                res[f"{label}_{stat}"] = ts_array.max()
            elif stat == "min":
                res[f"{label}_{stat}"] = ts_array.min()
            elif stat == "kurt":
                res[f"{label}_{stat}"] = stats.kurtosis(ts_array)
            elif stat == "skew":
                res[f"{label}_{stat}"] = stats.skew(ts_array)
            elif stat == "std":
                res[f"{label}_{stat}"] = np.std(ts_array)
            else:
                res[f"{label}_{stat}"] = np.percentile(ts_array, 100 - stat)
        except Exception as e:
            logger.error(f"Error calculating {stat} for {label}: {str(e)}")
            res[f"{label}_{stat}"] = np.nan
    return res


def mosqito_metric_1ch(
    s,
    metric: str,
    statistics: Tuple[Union[int, str]] = (
        5,
        10,
        50,
        90,
        95,
        "avg",
        "max",
        "min",
        "kurt",
        "skew",
    ),
    label: Optional[str] = None,
    as_df: bool = False,
    return_time_series: bool = False,
    func_args: Dict = {},
) -> Union[Dict, pd.DataFrame]:
    """
    Calculate a MoSQITo psychoacoustic metric for a single channel signal.

    Parameters
    ----------
    s : Signal
        Single channel signal object to analyze.
    metric : str
        Name of the metric to calculate. Options are "loudness_zwtv",
        "roughness_dw", "sharpness_din_from_loudness", "sharpness_din_perseg",
        or "sharpness_din_tv".
    statistics : Tuple[Union[int, str], ...], optional
        Statistics to calculate on the metric results.
    label : str, optional
        Label to use for the metric in the results. If None, uses a default label.
    as_df : bool, optional
        If True, return results as a pandas DataFrame. Otherwise, return a dictionary.
    return_time_series : bool, optional
        If True, include the full time series in the results.
    func_args : dict, optional
        Additional arguments to pass to the underlying MoSQITo function.

    Returns
    -------
    Union[dict, pd.DataFrame]
        Results of the metric calculation and statistics.

    Raises
    ------
    ValueError
        If the input signal is not single-channel or if an unrecognized metric is specified.

    Examples
    --------
    >>> # xdoctest: +SKIP
    >>> from soundscapy.audio import Binaural
    >>> signal = Binaural.from_wav("audio.wav", resample=480000)
    >>> results = mosqito_metric_1ch(signal[0], "loudness_zwtv", as_df=True)
    """
    logger.debug(f"Calculating MoSQITo metric: {metric}")

    # Checks and warnings
    if s.channels != 1:
        logger.error("Signal must be single channel")
        raise ValueError("Signal must be single channel")
    try:
        label = label or DEFAULT_LABELS[metric]
    except KeyError as e:
        logger.error(f"Metric {metric} not recognized")
        raise ValueError(f"Metric {metric} not recognized.") from e
    if as_df and return_time_series:
        logger.warning(
            "Cannot return both a dataframe and time series. Returning dataframe only."
        )
        return_time_series = False

    # Start the calc
    res = {}
    try:
        if metric == "loudness_zwtv":
            N, N_spec, bark_axis, time_axis = loudness_zwtv(s, s.fs, **func_args)
            res = _stat_calcs(label, N, res, statistics)
            if return_time_series:
                res[f"{label}_ts"] = (time_axis, N)
        elif metric == "roughness_dw":
            R, R_spec, bark_axis, time_axis = roughness_dw(s, s.fs, **func_args)
            res = _stat_calcs(label, R, res, statistics)
            if return_time_series:
                res[f"{label}_ts"] = (time_axis, R)
        elif metric == "sharpness_din_from_loudness":
            field_type = func_args.get("field_type", "free")
            N, N_spec, bark_axis, time_axis = loudness_zwtv(
                s, s.fs, field_type=field_type
            )
            res = _stat_calcs("N", N, res, statistics)
            if return_time_series:
                res["N_ts"] = time_axis, N

            func_args.pop("field_type", None)
            S = sharpness_din_from_loudness(N, N_spec, **func_args)
            res = _stat_calcs(label, S, res, statistics)
            if return_time_series:
                res[f"{label}_ts"] = (time_axis, S)
        elif metric == "sharpness_din_perseg":
            S, time_axis = sharpness_din_perseg(s, s.fs, **func_args)
            res = _stat_calcs(label, S, res, statistics)
            if return_time_series:
                res[f"{label}_ts"] = (time_axis, S)
        elif metric == "sharpness_din_tv":
            S, time_axis = sharpness_din_tv(s, s.fs, **func_args)
            res = _stat_calcs(label, S, res, statistics)
            if return_time_series:
                res[f"{label}_ts"] = (time_axis, S)
        else:
            logger.error(f"Metric {metric} not recognized")
            raise ValueError(f"Metric {metric} not recognized.")
    except Exception as e:
        logger.error(f"Error calculating {metric}: {str(e)}")
        raise

    # Return the results in the requested format
    if not as_df:
        return res
    try:
        rec = s.recording
        return pd.DataFrame(res, index=[rec])
    except AttributeError:
        return pd.DataFrame(res, index=[0])


def maad_metric_1ch(s, metric: str, as_df: bool = False, func_args={}):
    """
    Run a metric from the scikit-maad library (or suite of indices) on a single channel signal.

    Currently only supports running all of the alpha indices at once.

    Parameters
    ----------
    s : Signal or Binaural (single channel)
        Single channel signal to calculate the alpha indices for.
    metric : {"all_temporal_alpha_indices", "all_spectral_alpha_indices"}
        Metric to calculate.
    as_df : bool, optional
        Whether to return a pandas DataFrame, by default False.
        If True, returns a MultiIndex Dataframe with ("Recording", "Channel") as the index.
    func_args : dict, optional
        Additional keyword arguments to pass to the metric function, by default {}.

    Returns
    -------
    dict or pd.DataFrame
        Dictionary of results if as_df is False, otherwise a pandas DataFrame.

    Raises
    ------
    ValueError
        If the signal is not single-channel or if an unrecognized metric is specified.

    See Also
    --------
    maad.features.all_spectral_alpha_indices
    maad.features.all_temporal_alpha_indices
    """
    logger.debug(f"Calculating MAAD metric: {metric}")

    # Checks and status
    if s.channels != 1:
        logger.error("Signal must be single channel")
        raise ValueError("Signal must be single channel")

    logger.debug(f"Calculating scikit-maad {metric}")

    # Start the calc
    try:
        if metric == "all_spectral_alpha_indices":
            Sxx, tn, fn, ext = spectrogram(s, s.fs, **func_args)
            res = all_spectral_alpha_indices(Sxx, tn, fn, extent=ext, **func_args)[0]
        elif metric == "all_temporal_alpha_indices":
            res = all_temporal_alpha_indices(s, s.fs, **func_args)
        else:
            logger.error(f"Metric {metric} not recognized")
            raise ValueError(f"Metric {metric} not recognized.")
    except Exception as e:
        logger.error(f"Error calculating {metric}: {str(e)}")
        raise

    if not as_df:
        return res.to_dict("records")[0]
    try:
        res["Recording"] = s.recording
        res.set_index(["Recording"], inplace=True)
        return res
    except AttributeError:
        return res


def pyacoustics_metric_1ch(
    s,
    metric: str,
    statistics: List[Union[int, str]] = (
        5,
        10,
        50,
        90,
        95,
        "avg",
        "max",
        "min",
        "kurt",
        "skew",
    ),
    label: str = None,
    as_df: bool = False,
    return_time_series: bool = False,
    func_args={},
):
    """
    Run a metric from the pyacoustics library on a single channel object.

    Parameters
    ----------
    s : Signal or Binaural (single channel slice)
        Single channel signal to calculate the metric for.
    metric : {"LZeq", "Leq", "LAeq", "LCeq", "SEL"}
        The metric to run.
    statistics : List[Union[int, str]], optional
        List of level statistics to calculate (e.g. L_5, L_90, etc).
    label : str, optional
        Label to use for the metric in the results dictionary.
        If None, will pull from default label for that metric given in DEFAULT_LABELS.
    as_df : bool, optional
        Whether to return a pandas DataFrame, by default False.
        If True, returns a MultiIndex Dataframe with ("Recording", "Channel") as the index.
    return_time_series : bool, optional
        Whether to return the time series of the metric, by default False.
        Cannot return time series if as_df is True.
    func_args : dict, optional
        Additional keyword arguments to pass to the metric function, by default {}.

    Returns
    -------
    dict or pd.DataFrame
        Dictionary of the calculated statistics or a pandas DataFrame.

    Raises
    ------
    ValueError
        If the signal is not single-channel or if an unrecognized metric is specified.

    See Also
    --------
    acoustics
    """
    logger.debug(f"Calculating pyacoustics metric: {metric}")

    if s.channels != 1:
        logger.error("Signal must be single channel")
        raise ValueError("Signal must be single channel")
    try:
        label = label or DEFAULT_LABELS[metric]
    except KeyError as e:
        logger.error(f"Metric {metric} not recognized")
        raise ValueError(f"Metric {metric} not recognized.") from e
    if as_df and return_time_series:
        logger.warning(
            "Cannot return both a dataframe and time series. Returning dataframe only."
        )

        return_time_series = False

    logger.debug(f"Calculating Python Acoustics: {metric} {statistics}")

    res = {}
    try:
        if metric in {"LZeq", "Leq", "LAeq", "LCeq"}:
            if metric in {"LZeq", "Leq"}:
                weighting = "Z"
            elif metric == "LAeq":
                weighting = "A"
            elif metric == "LCeq":
                weighting = "C"
            if "avg" in statistics or "mean" in statistics:
                stat = "avg" if "avg" in statistics else "mean"
                res[f"{label}"] = s.weigh(weighting).leq()
                statistics = list(statistics)
                statistics.remove(stat)
            if len(statistics) > 0:
                res = _stat_calcs(
                    label, s.weigh(weighting).levels(**func_args)[1], res, statistics
                )

            if return_time_series:
                res[f"{label}_ts"] = s.weigh(weighting).levels(**func_args)
        elif metric == "SEL":
            res[f"{label}"] = s.sound_exposure_level()
        else:
            logger.error(f"Metric {metric} not recognized")
            raise ValueError(f"Metric {metric} not recognized.")
    except Exception as e:
        logger.error(f"Error calculating {metric}: {str(e)}")
        raise

    if not as_df:
        return res
    try:
        rec = s.recording
        return pd.DataFrame(res, index=[rec])
    except AttributeError:
        return pd.DataFrame(res, index=[0])


def pyacoustics_metric_2ch(
    b,
    metric: str,
    statistics: Union[Tuple, List] = (
        5,
        10,
        50,
        90,
        95,
        "avg",
        "max",
        "min",
        "kurt",
        "skew",
    ),
    label: str = None,
    channel_names: Tuple[str, str] = ("Left", "Right"),
    as_df: bool = False,
    return_time_series: bool = False,
    func_args={},
):
    """
    Run a metric from the python acoustics library on a Binaural object.

    Parameters
    ----------
    b : Binaural
        Binaural signal to calculate the metric for.
    metric : {"LZeq", "Leq", "LAeq", "LCeq", "SEL"}
        The metric to run.
    statistics : tuple or list, optional
        List of level statistics to calculate (e.g. L_5, L_90, etc).
    label : str, optional
        Label to use for the metric in the results dictionary.
        If None, will pull from default label for that metric given in DEFAULT_LABELS.
    channel_names : tuple of str, optional
        Custom names for the channels, by default ("Left", "Right").
    as_df : bool, optional
        Whether to return a pandas DataFrame, by default False.
        If True, returns a MultiIndex Dataframe with ("Recording", "Channel") as the index.
    return_time_series : bool, optional
        Whether to return the time series of the metric, by default False.
        Cannot return time series if as_df is True.
    func_args : dict, optional
        Arguments to pass to the metric function, by default {}.

    Returns
    -------
    dict or pd.DataFrame
        Dictionary of results if as_df is False, otherwise a pandas DataFrame.

    Raises
    ------
    ValueError
        If the input signal is not 2-channel.

    See Also
    --------
    pyacoustics_metric_1ch
    """
    logger.debug(f"Calculating pyacoustics metric for 2 channels: {metric}")

    if b.channels != 2:
        logger.error("Must be 2 channel signal. Use `pyacoustics_metric_1ch` instead.")
        raise ValueError(
            "Must be 2 channel signal. Use `pyacoustics_metric_1ch instead`."
        )

    logger.debug(f"Calculating Python Acoustics metrics: {metric}")

    try:
        res_l = pyacoustics_metric_1ch(
            b[0],
            metric,
            statistics,
            label,
            as_df=False,
            return_time_series=return_time_series,
            func_args=func_args,
        )

        res_r = pyacoustics_metric_1ch(
            b[1],
            metric,
            statistics,
            label,
            as_df=False,
            return_time_series=return_time_series,
            func_args=func_args,
        )

        res = {channel_names[0]: res_l, channel_names[1]: res_r}
    except Exception as e:
        logger.error(f"Error calculating {metric} for 2 channels: {str(e)}")
        raise

    if not as_df:
        return res
    try:
        rec = b.recording
    except AttributeError:
        rec = 0
    df = pd.DataFrame.from_dict(res, orient="index")
    df["Recording"] = rec
    df["Channel"] = df.index
    df.set_index(["Recording", "Channel"], inplace=True)
    return df


def _parallel_mosqito_metric_2ch(
    b,
    metric: str,
    statistics: Union[Tuple, List] = (
        5,
        10,
        50,
        90,
        95,
        "avg",
        "max",
        "min",
        "kurt",
        "skew",
    ),
    label: str = None,
    channel_names: Tuple[str, str] = ("Left", "Right"),
    return_time_series: bool = False,
    func_args={},
):
    """
    Run a metric from the mosqito library on a Binaural object in parallel.

    Parameters
    ----------
    b : Binaural
        Binaural signal to calculate the metric for.
    metric : str
        The metric to run.
    statistics : tuple or list, optional
        List of level statistics to calculate (e.g. L_5, L_90, etc).
    label : str, optional
        Label to use for the metric in the results dictionary.
        If None, will pull from default label for that metric given in DEFAULT_LABELS.
    channel_names : tuple of str, optional
        Custom names for the channels, by default ("Left", "Right").
    return_time_series : bool, optional
        Whether to return the time series of the metric, by default False.
    func_args : dict, optional
        Arguments to pass to the metric function, by default {}.

    Returns
    -------
    dict
        Dictionary of results for both channels.

    See Also
    --------
    mosqito_metric_1ch
    """
    logger.debug(f"Calculating MoSQITo metric in parallel: {metric}")

    pool = mp.Pool(mp.cpu_count())
    try:
        result_objects = pool.starmap(
            mosqito_metric_1ch,
            [
                (
                    b[i],
                    metric,
                    statistics,
                    label,
                    False,
                    return_time_series,
                    func_args,
                )
                for i in [0, 1]
            ],
        )

        pool.close()
        return {channel: result_objects[i] for i, channel in enumerate(channel_names)}
    except Exception as e:
        logger.error(f"Error in parallel MoSQITo calculation: {str(e)}")
        pool.close()
        raise
    finally:
        pool.join()


def mosqito_metric_2ch(
    b,
    metric: str,
    statistics: Union[Tuple, List] = (
        5,
        10,
        50,
        90,
        95,
        "avg",
        "max",
        "min",
        "kurt",
        "skew",
    ),
    label: str = None,
    channel_names: Tuple[str, str] = ("Left", "Right"),
    as_df: bool = False,
    return_time_series: bool = False,
    parallel: bool = True,
    func_args={},
):
    """
    Calculate metrics from MoSQITo for a two-channel signal with optional parallel processing.

    Parameters
    ----------
    b : Binaural
        Binaural signal to calculate the sound quality indices for.
    metric : {"loudness_zwtv", "sharpness_din_from_loudness", "sharpness_din_perseg",
    "sharpness_din_tv", "roughness_dw"}
        Metric to calculate.
    statistics : tuple or list, optional
        List of level statistics to calculate (e.g. L_5, L_90, etc.).
    label : str, optional
        Label to use for the metric in the results dictionary.
        If None, will pull from default label for that metric given in DEFAULT_LABELS.
    channel_names : tuple of str, optional
        Custom names for the channels, by default ("Left", "Right").
    as_df : bool, optional
        Whether to return a pandas DataFrame, by default False.
        If True, returns a MultiIndex Dataframe with ("Recording", "Channel") as the index.
    return_time_series : bool, optional
        Whether to return the time series of the metric, by default False.
        Only works for metrics that return a time series array.
        Cannot be returned in a dataframe.
    parallel : bool, optional
        Whether to process channels in parallel, by default True.
    func_args : dict, optional
        Additional arguments to pass to the metric function, by default {}.

    Returns
    -------
    dict or pd.DataFrame
        Dictionary of results if as_df is False, otherwise a pandas DataFrame.

    Raises
    ------
    ValueError
        If the input signal is not 2-channel.
    """
    logger.debug(f"Calculating MoSQITo metric for 2 channels: {metric}")

    if b.channels != 2:
        logger.error("Must be 2 channel signal. Use `mosqito_metric_1ch` instead.")
        raise ValueError("Must be 2 channel signal. Use `mosqito_metric_1ch` instead.")

    if metric == "sharpness_din_from_loudness":
        logger.debug(
            "Calculating MoSQITo metrics: `sharpness_din` from `loudness_zwtv`"
        )
    else:
        logger.debug(f"Calculating MoSQITo metric: {metric}")

    try:
        if parallel:
            with concurrent.futures.ThreadPoolExecutor(max_workers=2) as executor:
                future_l = executor.submit(
                    mosqito_metric_1ch,
                    b[0],
                    metric,
                    statistics,
                    label,
                    False,
                    return_time_series,
                    func_args,
                )
                future_r = executor.submit(
                    mosqito_metric_1ch,
                    b[1],
                    metric,
                    statistics,
                    label,
                    False,
                    return_time_series,
                    func_args,
                )
                res_l = future_l.result()
                res_r = future_r.result()
        else:
            res_l = mosqito_metric_1ch(
                b[0],
                metric,
                statistics,
                label,
                False,
                return_time_series,
                func_args,
            )
            res_r = mosqito_metric_1ch(
                b[1],
                metric,
                statistics,
                label,
                False,
                return_time_series,
                func_args,
            )

        res = {channel_names[0]: res_l, channel_names[1]: res_r}
    except Exception as e:
        logger.error(
            f"Error calculating MoSQITo metric {metric} for 2 channels: {str(e)}"
        )
        raise

    if not as_df:
        return res
    try:
        rec = b.recording
    except AttributeError:
        rec = 0
    df = pd.DataFrame.from_dict(res, orient="index")
    df["Recording"] = rec
    df["Channel"] = df.index
    df.set_index(["Recording", "Channel"], inplace=True)
    return df


def maad_metric_2ch(
    b,
    metric: str,
    channel_names: Tuple[str, str] = ("Left", "Right"),
    as_df: bool = False,
    func_args={},
):
    """
    Run a metric from the scikit-maad library (or suite of indices) on a binaural signal.

    Currently only supports running all the alpha indices at once.

    Parameters
    ----------
    b : Binaural
        Binaural signal to calculate the alpha indices for.
    metric : {"all_temporal_alpha_indices", "all_spectral_alpha_indices"}
        Metric to calculate.
    channel_names : tuple of str, optional
        Custom names for the channels, by default ("Left", "Right").
    as_df : bool, optional
        Whether to return a pandas DataFrame, by default False.
        If True, returns a MultiIndex Dataframe with ("Recording", "Channel") as the index.
    func_args : dict, optional
        Additional arguments to pass to the metric function, by default {}.

    Returns
    -------
    dict or pd.DataFrame
        Dictionary of results if as_df is False, otherwise a pandas DataFrame.

    Raises
    ------
    ValueError
        If the input signal is not 2-channel or if an unrecognized metric is specified.

    See Also
    --------
    scikit-maad library
    maad_metric_1ch
    """
    logger.debug(f"Calculating MAAD metric for 2 channels: {metric}")

    if b.channels != 2:
        logger.error("Must be 2 channel signal. Use `maad_metric_1ch` instead.")
        raise ValueError("Must be 2 channel signal. Use `maad_metric_1ch` instead.")

    logger.debug(f"Calculating scikit-maad {metric}")

    try:
        res_l = maad_metric_1ch(b[0], metric, as_df=False)
        res_r = maad_metric_1ch(b[1], metric, as_df=False)
        res = {channel_names[0]: res_l, channel_names[1]: res_r}
    except Exception as e:
        logger.error(f"Error calculating MAAD metric {metric} for 2 channels: {str(e)}")
        raise

    if not as_df:
        return res
    try:
        rec = b.recording
    except AttributeError:
        rec = 0
    df = pd.DataFrame.from_dict(res, orient="index")
    df["Recording"] = rec
    df["Channel"] = df.index
    df.set_index(["Recording", "Channel"], inplace=True)
    return df


# Analysis dataframe functions
def prep_multiindex_df(dictionary: dict, label: str = "Leq", incl_metric: bool = True):
    """
    Prepare a MultiIndex dataframe from a dictionary of results.

    Parameters
    ----------
    dictionary : dict
        Dict of results with recording name as key, channels {"Left", "Right"} as second key,
        and Leq metric as value.
    label : str, optional
        Name of metric included, by default "Leq".
    incl_metric : bool, optional
        Whether to include the metric value in the resulting dataframe, by default True.
        If False, will only set up the DataFrame with the proper MultiIndex.

    Returns
    -------
    pd.DataFrame
        Index includes "Recording" and "Channel" with a column for each index if `incl_metric`.

    Raises
    ------
    ValueError
        If the input dictionary is not in the expected format.
    """
    logger.info("Preparing MultiIndex DataFrame")
    try:
        new_dict = {}
        for outerKey, innerDict in dictionary.items():
            for innerKey, values in innerDict.items():
                new_dict[(outerKey, innerKey)] = values
        idx = pd.MultiIndex.from_tuples(new_dict.keys())
        df = pd.DataFrame(new_dict.values(), index=idx, columns=[label])
        df.index.names = ["Recording", "Channel"]
        if not incl_metric:
            df = df.drop(columns=[label])
        logger.debug("MultiIndex DataFrame prepared successfully")
        return df
    except Exception as e:
        logger.error(f"Error preparing MultiIndex DataFrame: {str(e)}")
        raise ValueError("Invalid input dictionary format") from e


def add_results(results_df: pd.DataFrame, metric_results: pd.DataFrame):
    """
    Add results to MultiIndex dataframe.

    Parameters
    ----------
    results_df : pd.DataFrame
        MultiIndex dataframe to add results to.
    metric_results : pd.DataFrame
        MultiIndex dataframe of results to add.

    Returns
    -------
    pd.DataFrame
        Index includes "Recording" and "Channel" with a column for each index.

    Raises
    ------
    ValueError
        If the input DataFrames are not in the expected format.
    """
    logger.info("Adding results to MultiIndex DataFrame")
    try:
        # TODO: Add check for whether all of the recordings have rows in the dataframe
        # If not, add new rows first

        if not set(metric_results.columns).issubset(set(results_df.columns)):
            # Check if results_df already has the columns in results
            results_df = results_df.join(metric_results)
        else:
            results_df.update(metric_results, errors="ignore")
        logger.debug("Results added successfully")
        return results_df
    except Exception as e:
        logger.error(f"Error adding results to DataFrame: {str(e)}")
        raise ValueError("Invalid input DataFrame format") from e


def process_all_metrics(
    b, analysis_settings: AnalysisSettings, parallel: bool = True
) -> pd.DataFrame:
    """
    Process all metrics specified in the analysis settings for a binaural signal.

    This function runs through all enabled metrics in the provided analysis settings,
    computes them for the given binaural signal, and compiles the results into a single DataFrame.

    Parameters
    ----------
    b : Binaural
        Binaural signal object to process.
    analysis_settings : AnalysisSettings
        Configuration object specifying which metrics to run and their parameters.
    parallel : bool, optional
        If True, run applicable calculations in parallel. Defaults to True.

    Returns
    -------
    pd.DataFrame
        A MultiIndex DataFrame containing results from all processed metrics.
        The index includes "Recording" and "Channel" levels.

    Raises
    ------
    ValueError
        If there's an error processing any of the metrics.

    Notes
    -----
    The parallel option primarily affects the MoSQITo metrics. Other metrics may not
    benefit from parallelization.

    Examples
    --------
    >>> # xdoctest: +SKIP
    >>> from soundscapy.audio import Binaural
    >>> from soundscapy import AnalysisSettings
    >>> signal = Binaural.from_wav("audio.wav", resample=480000)
    >>> settings = AnalysisSettings.from_yaml("settings.yaml")
    >>> results = process_all_metrics(signal,settings)
    """
    logger.info(f"Processing all metrics for {b.recording}")
    logger.debug(f"Parallel processing: {parallel}")

    idx = pd.MultiIndex.from_tuples(((b.recording, "Left"), (b.recording, "Right")))
    results_df = pd.DataFrame(index=idx)
    results_df.index.names = ["Recording", "Channel"]

    try:
        for (
            library,
            metrics_settings,
        ) in analysis_settings.get_enabled_metrics().items():
            for metric in metrics_settings.keys():
                logger.debug(f"Processing {library} metric: {metric}")
                if library == "PythonAcoustics":
                    results_df = pd.concat(
                        (
                            results_df,
                            b.pyacoustics_metric(
                                metric, metric_settings=metrics_settings[metric]
                            ),
                        ),
                        axis=1,
                    )
                elif library == "MoSQITo":
                    results_df = pd.concat(
                        (
                            results_df,
                            b.mosqito_metric(
                                metric,
                                parallel=parallel,
                                metric_settings=metrics_settings[metric],
                            ),
                        ),
                        axis=1,
                    )
                elif library == "scikit-maad" or library == "scikit_maad":
                    results_df = pd.concat(
                        (
                            results_df,
                            b.maad_metric(
                                metric, metric_settings=metrics_settings[metric]
                            ),
                        ),
                        axis=1,
                    )
        logger.info("All metrics processed successfully")
        return results_df
    except Exception as e:
        logger.error(f"Error processing metrics: {str(e)}")
        raise ValueError("Error processing metrics") from e


# Add any additional helper functions or constants here if needed

if __name__ == "__main__":
    # Add any script-level code or examples here
    pass



"""
Script to generate baseline images for pytest-mpl comparisons.
"""

import os

import matplotlib.pyplot as plt
from numpy import random

from soundscapy.plotting import Backend, density_plot, scatter_plot
from soundscapy.surveys.processing import simulation

# Set random seed for reproducibility
random.seed(42)

# Create directory for baseline images if it doesn't exist
BASELINE_DIR = "/Users/mitch/Documents/GitHub/Soundscapy/test/baseline"
os.makedirs(BASELINE_DIR, exist_ok=True)

# Generate sample data
sample_data = simulation(n=100, incl_iso_coords=True)

# Generate and save scatter plot
scatter_ax = scatter_plot(sample_data, backend=Backend.SEABORN)
scatter_ax.figure.savefig(os.path.join(BASELINE_DIR, "test_scatter_plot.png"))
plt.close(scatter_ax.figure)

# Generate and save density plot
density_ax = density_plot(sample_data, backend=Backend.SEABORN)
density_ax.figure.savefig(os.path.join(BASELINE_DIR, "test_density_plot.png"))
plt.close(density_ax.figure)

print("Baseline images generated successfully.")



import numpy as np
import pandas as pd
import pytest
from pytest import approx

from soundscapy.surveys.processing import (
    add_iso_coords,
    calculate_iso_coords,
    likert_data_quality,
    simulation,
    ssm_metrics,
)
from soundscapy.surveys.survey_utils import (
    LANGUAGE_ANGLES,
    PAQ_IDS,
    rename_paqs,
    return_paqs,
)


@pytest.fixture
def basic_test_df():
    return pd.DataFrame(
        {
            "RecordID": ["EX1", "EX2", "EX3", "EX4"],
            "GroupID": ["A", "A", "B", "B"],
            "PAQ1": [4, 2, 3, 5],
            "PAQ2": [4, 3, 5, 2],
            "PAQ3": [4, 5, 2, 3],
            "PAQ4": [2, 5, 3, 4],
            "PAQ5": [1, 5, 4, 2],
            "PAQ6": [3, 5, 2, 4],
            "PAQ7": [3, 3, 5, 1],
            "PAQ8": [4, 1, 2, 5],
        }
    )


@pytest.fixture
def name_test_df():
    return pd.DataFrame(
        {
            "RecordID": ["EX1", "EX2"],
            "pleasant": [4, 2],
            "vibrant": [4, 3],
            "eventful": [4, 5],
            "chaotic": [2, 5],
            "annoying": [1, 5],
            "monotonous": [3, 5],
            "uneventful": [3, 3],
            "calm": [4, 1],
        }
    )


class TestSurveyUtils:
    def test_return_paqs(self, basic_test_df):
        result = return_paqs(basic_test_df)
        assert list(result.columns) == ["RecordID", "GroupID"] + PAQ_IDS

        result = return_paqs(basic_test_df, incl_ids=False, other_cols=["GroupID"])
        assert list(result.columns) == PAQ_IDS + ["GroupID"]

    def test_rename_paqs(self, name_test_df):
        result = rename_paqs(name_test_df)
        assert list(result.columns) == ["RecordID"] + PAQ_IDS

        custom_aliases = {
            "pl": "PAQ1",
            "ch": "PAQ4",
            "ca": "PAQ8",
            "v": "PAQ2",
            "ev": "PAQ3",
            "un": "PAQ7",
            "ann": "PAQ5",
            "m": "PAQ6",
        }
        df = pd.DataFrame(
            {
                "RecordID": ["EX1", "EX2"],
                "pl": [4, 2],
                "ch": [4, 3],
                "ca": [4, 5],
                "v": [2, 5],
                "ev": [1, 5],
                "un": [3, 5],
                "ann": [3, 3],
                "m": [4, 1],
            }
        )
        result = rename_paqs(df, paq_aliases=custom_aliases)
        assert list(result.columns) == ["RecordID"] + list(custom_aliases.values())


class TestISOCoordinates:
    def test_calculate_iso_coords(self, basic_test_df):
        iso_pleasant, iso_eventful = calculate_iso_coords(basic_test_df)
        assert isinstance(iso_pleasant, pd.Series)
        assert isinstance(iso_eventful, pd.Series)
        assert len(iso_pleasant) == len(basic_test_df)
        assert len(iso_eventful) == len(basic_test_df)
        assert iso_pleasant.iloc[0] == approx(0.53, abs=0.05)
        assert iso_eventful.iloc[0] == approx(0.03, abs=0.05)

    def test_add_iso_coords(self, basic_test_df):
        df = add_iso_coords(basic_test_df)
        assert "ISOPleasant" in df.columns and "ISOEventful" in df.columns

        with pytest.raises(Warning):
            df = add_iso_coords(df, overwrite=False)

        df = add_iso_coords(df, names=("pl_test", "ev_test"), overwrite=True)
        assert "pl_test" in df.columns and "ev_test" in df.columns

    def test_iso_coords_with_language_angles(self, basic_test_df):
        iso_pleasant_eng, iso_eventful_eng = calculate_iso_coords(
            basic_test_df, angles=LANGUAGE_ANGLES["eng"]
        )
        iso_pleasant_cmn, iso_eventful_cmn = calculate_iso_coords(
            basic_test_df, angles=LANGUAGE_ANGLES["cmn"]
        )

        assert iso_pleasant_eng.iloc[0] == approx(0.66, abs=0.01)
        assert iso_eventful_eng.iloc[0] == approx(0.14, abs=0.01)
        assert iso_pleasant_cmn.iloc[0] == approx(0.41, abs=0.01)
        assert iso_eventful_cmn.iloc[0] == approx(-0.09, abs=0.01)


class TestDataQuality:
    def test_likert_data_quality(self, basic_test_df):
        result = likert_data_quality(basic_test_df)
        assert result is None  # Assuming all data in basic_test_df is valid

    def test_likert_data_quality_with_invalid_data(self):
        invalid_df = pd.DataFrame(
            {
                "PAQ1": [6, 2, 3, 5],
                "PAQ2": [4, 3, 5, 2],
                "PAQ3": [4, 5, 2, 3],
                "PAQ4": [2, 5, 3, 4],
                "PAQ5": [1, 5, 4, 2],
                "PAQ6": [3, 5, 2, 4],
                "PAQ7": [3, 3, 5, 1],
                "PAQ8": [4, 1, 2, 5],
            }
        )
        result = likert_data_quality(invalid_df)
        assert result == [0]

    def test_likert_data_quality_with_nan(self):
        nan_df = pd.DataFrame(
            {
                "PAQ1": [4, 2, np.nan, 5],
                "PAQ2": [4, 3, 5, 2],
                "PAQ3": [4, 5, 2, 3],
                "PAQ4": [2, 5, 3, 4],
                "PAQ5": [1, 5, 4, 2],
                "PAQ6": [3, 5, 2, 4],
                "PAQ7": [3, 3, 5, 1],
                "PAQ8": [4, 1, 2, 5],
            }
        )
        result = likert_data_quality(nan_df)
        assert result == [2]

        result_allow_na = likert_data_quality(nan_df, allow_na=True)
        assert result_allow_na is None


class TestSimulation:
    def test_simulation(self):
        df = simulation(n=200)
        assert df.shape == (200, 8)
        assert list(df.columns) == PAQ_IDS
        assert df.min().min() >= 1 and df.max().max() <= 5

    def test_simulation_with_iso_coords(self):
        df = simulation(n=200, incl_iso_coords=True)
        assert df.shape == (200, 10)
        assert list(df.columns) == PAQ_IDS + ["ISOPleasant", "ISOEventful"]
        assert not df["ISOPleasant"].isna().any() and not df["ISOEventful"].isna().any()


class TestSSMMetrics:
    @pytest.fixture
    def ssm_test_df(self):
        return pd.DataFrame(
            {
                "PAQ1": [4, 2, 3, 5],
                "PAQ2": [4, 3, 5, 2],
                "PAQ3": [4, 5, 2, 3],
                "PAQ4": [2, 5, 3, 4],
                "PAQ5": [1, 5, 4, 2],
                "PAQ6": [3, 5, 2, 4],
                "PAQ7": [3, 3, 5, 1],
                "PAQ8": [4, 1, 2, 5],
            }
        )

    def test_ssm_metrics_cosine(self, ssm_test_df):
        with pytest.warns(PendingDeprecationWarning):
            result = ssm_metrics(ssm_test_df, method="cosine")
            assert isinstance(result, pd.DataFrame)
            assert list(result.columns) == [
                "amplitude",
                "angle",
                "elevation",
                "displacement",
                "r_squared",
            ]
            assert len(result) == len(ssm_test_df)
            assert (result["r_squared"] >= 0).all() and (result["r_squared"] <= 1).all()

    def test_ssm_metrics_polar(self, ssm_test_df):
        with pytest.warns(PendingDeprecationWarning):
            result = ssm_metrics(ssm_test_df, method="polar")
            assert isinstance(result, pd.DataFrame)
            assert list(result.columns) == [
                "amplitude",
                "angle",
                "elevation",
                "displacement",
                "r_squared",
            ]
            assert len(result) == len(ssm_test_df)
            assert (result["displacement"] == 0).all()
            assert (result["r_squared"] == 1).all()

    def test_ssm_metrics_invalid_method(self, ssm_test_df):
        with pytest.warns(PendingDeprecationWarning):
            with pytest.raises(ValueError):
                ssm_metrics(ssm_test_df, method="invalid")

    def test_ssm_metrics_missing_columns(self):
        with pytest.warns(PendingDeprecationWarning):
            invalid_df = pd.DataFrame({"PAQ1": [1, 2, 3], "PAQ2": [4, 5, 6]})
            with pytest.raises(ValueError):
                ssm_metrics(invalid_df)


class TestIntegration:
    def test_end_to_end_workflow(self):
        # Simulate data
        df = simulation(n=100, incl_iso_coords=True)

        # Check data quality
        invalid_indices = likert_data_quality(df)
        assert invalid_indices is None

        # Calculate SSM metrics
        ssm_result = ssm_metrics(df)

        # Verify results
        assert df.shape == (100, 10)
        assert ssm_result.shape == (100, 5)
        assert not df["ISOPleasant"].isna().any() and not df["ISOEventful"].isna().any()
        assert not ssm_result.isna().any().any()


if __name__ == "__main__":
    pytest.main()



"""
Module for handling the Soundscape Attributes Translation Project (SATP) database.

This module provides functions for loading and processing data from the
Soundscape Attributes Translation Project database. It includes utilities
for data retrieval from Zenodo and basic data loading operations.

Examples
--------
>>> import soundscapy.databases.satp as satp  # doctest: +SKIP
>>> df = satp.load_zenodo()  # doctest: +SKIP
>>> isinstance(df, pd.DataFrame)  # doctest: +SKIP
True
>>> 'Language' in df.columns  # doctest: +SKIP
True
>>> participants = satp.load_participants()  # doctest: +SKIP
>>> isinstance(participants, pd.DataFrame)  # doctest: +SKIP
True
>>> 'Country' in participants.columns  # doctest: +SKIP
True
"""

import pandas as pd
from loguru import logger


def _url_fetch(version: str) -> str:
    """
    Return the URL to fetch the SATP dataset from Zenodo.

    Parameters
    ----------
    version : str
        Version of the dataset to load.

    Returns
    -------
    str
        URL to fetch the SATP dataset from Zenodo.

    Raises
    ------
    ValueError
        If an invalid version is specified.

    Examples
    --------
    >>> _url_fetch("latest")
    'https://zenodo.org/record/7143599/files/SATP%20Dataset%20v1.2.xlsx'
    >>> _url_fetch("v1.2.1")
    'https://zenodo.org/record/7143599/files/SATP%20Dataset%20v1.2.xlsx'
    >>> _url_fetch("invalid")
    Traceback (most recent call last):
        ...
    ValueError: Invalid version. Should be either 'latest', 'v1.2.1', or 'v1.2'.
    """
    if version.lower() not in ["latest", "v1.2.1", "v1.2"]:
        raise ValueError(
            "Invalid version. Should be either 'latest', 'v1.2.1', or 'v1.2'."
        )

    version = "v1.2.1" if version.lower() == "latest" else version.lower()
    url = "https://zenodo.org/record/7143599/files/SATP%20Dataset%20v1.2.xlsx"

    logger.debug(f"Fetching SATP dataset URL for version: {version}")
    return url


def load_zenodo(version: str = "latest") -> pd.DataFrame:
    """
    Load the SATP dataset from Zenodo.

    Parameters
    ----------
    version : str, optional
        Version of the dataset to load. The default is "latest".

    Returns
    -------
    pd.DataFrame
        DataFrame containing the SATP dataset.
    """
    url = _url_fetch(version)
    df = pd.read_excel(url, engine="openpyxl", sheet_name="Main Merge")
    logger.info(f"Loaded SATP dataset version {version} from Zenodo")
    return df


def load_participants(version: str = "latest") -> pd.DataFrame:
    """
    Load the SATP participants dataset from Zenodo.

    Parameters
    ----------
    version : str, optional
        Version of the dataset to load. The default is "latest".

    Returns
    -------
    pd.DataFrame
        DataFrame containing the SATP participants dataset.
    """
    url = _url_fetch(version)
    df = pd.read_excel(url, engine="openpyxl", sheet_name="Participants")
    df = df.drop(columns=["Unnamed: 3", "Unnamed: 4"])
    logger.info(f"Loaded SATP participants dataset version {version} from Zenodo")
    return df


if __name__ == "__main__":
    import xdoctest

    xdoctest.doctest_module(__file__)



import pytest
import os


def test_optional_dependency_groups_defined():
    """Test that OPTIONAL_DEPENDENCIES has expected structure."""
    from soundscapy._optionals import OPTIONAL_DEPENDENCIES

    assert "audio" in OPTIONAL_DEPENDENCIES

    group = OPTIONAL_DEPENDENCIES["audio"]
    assert "packages" in group
    assert "install" in group
    assert "description" in group

    assert isinstance(group["packages"], tuple)
    assert len(group["packages"]) > 0
    assert all(isinstance(pkg, str) for pkg in group["packages"])

    assert group["install"] == "soundscapy[audio]"
    assert isinstance(group["description"], str)


@pytest.mark.skipif(
    os.environ.get("AUDIO_DEPS") == "1",
    reason="Test requires audio dependencies to be missing",
)
def test_require_dependencies_missing():
    """Test behavior when dependencies are missing."""
    from soundscapy._optionals import require_dependencies

    with pytest.raises(ImportError) as exc_info:
        require_dependencies("audio")

    assert "Install with:" in str(exc_info.value)
    assert "pip install soundscapy[audio]" in str(exc_info.value)


@pytest.mark.optional_deps("audio")
def test_require_dependencies_present():
    """Test behavior when dependencies are present."""
    from soundscapy._optionals import require_dependencies

    packages = require_dependencies("audio")
    assert isinstance(packages, dict)
    assert all(pkg in packages for pkg in ["mosqito", "maad", "acoustics"])
    assert all(packages[pkg] is not None for pkg in packages)


def test_optional_import_behavior():
    """Test top-level optional component import behavior."""
    import importlib
    import soundscapy
    from soundscapy._optionals import OPTIONAL_IMPORTS

    # Test that optional components are listed in __all__
    for name in OPTIONAL_IMPORTS:
        assert name in soundscapy.__all__

    # Test import behavior
    def has_audio_deps():
        """Check if audio dependencies are available."""
        try:
            importlib.import_module("mosqito")
            importlib.import_module("maad")
            importlib.import_module("acoustics")
            return True
        except ImportError:
            return False

    # Try importing Binaural as an example component
    if has_audio_deps():
        # Should succeed when dependencies are available
        from soundscapy import Binaural

        assert hasattr(Binaural, "__module__")
    else:
        # Should raise helpful ImportError when dependencies missing
        import pytest

        with pytest.raises(ImportError) as exc_info:
            from soundscapy import Binaural
        assert "audio analysis functionality" in str(exc_info.value)
        assert "pip install soundscapy[audio]" in str(exc_info.value)


def test_invalid_group():
    """Test behavior with invalid dependency group."""
    from soundscapy._optionals import require_dependencies

    with pytest.raises(KeyError) as exc_info:
        require_dependencies("nonexistent_group")
    assert "Unknown dependency group" in str(exc_info.value)



"""
Core utility functions for processing soundscape survey data.

This module contains fundamental functions and constants used across
the soundscapy package for handling and analyzing soundscape survey data.
"""

from enum import Enum
from typing import Dict, List, Tuple, Union

import pandas as pd
from loguru import logger


class PAQ(Enum):
    """Enumeration of Perceptual Attribute Questions (PAQ) names and IDs."""

    PLEASANT = ("pleasant", "PAQ1")
    VIBRANT = ("vibrant", "PAQ2")
    EVENTFUL = ("eventful", "PAQ3")
    CHAOTIC = ("chaotic", "PAQ4")
    ANNOYING = ("annoying", "PAQ5")
    MONOTONOUS = ("monotonous", "PAQ6")
    UNEVENTFUL = ("uneventful", "PAQ7")
    CALM = ("calm", "PAQ8")

    def __init__(self, label: str, id: str):
        self.label = label
        self.id = id


PAQ_LABELS = [paq.label for paq in PAQ]
PAQ_IDS = [paq.id for paq in PAQ]

EQUAL_ANGLES = (0, 45, 90, 135, 180, 225, 270, 315)

# Language-specific angles for PAQs as defined in Aletta et. al. (2024)
LANGUAGE_ANGLES = {
    "eng": (0, 46, 94, 138, 177, 241, 275, 340),
    "arb": (0, 36, 45, 135, 167, 201, 242, 308),
    "cmn": (0, 18, 38, 154, 171, 196, 217, 318),
    "hrv": (0, 84, 93, 160, 173, 243, 273, 354),
    "nld": (0, 43, 111, 125, 174, 257, 307, 341),
    "deu": (0, 64, 97, 132, 182, 254, 282, 336),
    "ell": (0, 72, 86, 133, 161, 233, 267, 328),
    "ind": (0, 53, 104, 123, 139, 202, 284, 308),
    "ita": (0, 57, 104, 143, 170, 274, 285, 336),
    "spa": (0, 41, 103, 147, 174, 238, 279, 332),
    "swe": (0, 66, 87, 146, 175, 249, 275, 335),
    "tur": (0, 55, 97, 106, 157, 254, 289, 313),
}


def return_paqs(
    df: pd.DataFrame, incl_ids: bool = True, other_cols: List[str] = None
) -> pd.DataFrame:
    """
    Return only the PAQ columns from a DataFrame.

    Parameters
    ----------
    df : pd.DataFrame
        Input DataFrame containing PAQ data.
    incl_ids : bool, optional
        Whether to include ID columns (RecordID, GroupID, etc.), by default True.
    other_cols : List[str], optional
        Other columns to include in the output, by default None.

    Returns
    -------
    pd.DataFrame
        DataFrame containing only the PAQ columns and optionally ID and other specified columns.

    Examples
    --------
    >>> import pandas as pd
    >>> df = pd.DataFrame({
    ...     'RecordID': [1, 2],
    ...     'PAQ1': [4, 3],
    ...     'PAQ2': [2, 5],
    ...     'PAQ3': [1, 2],
    ...     'PAQ4': [3, 4],
    ...     'PAQ5': [5, 1],
    ...     'PAQ6': [2, 3],
    ...     'PAQ7': [4, 5],
    ...     'PAQ8': [1, 2],
    ...     'OtherCol': ['A', 'B']
    ... })
    >>> return_paqs(df)
       RecordID  PAQ1  PAQ2  PAQ3  PAQ4  PAQ5  PAQ6  PAQ7  PAQ8
    0         1     4     2     1     3     5     2     4     1
    1         2     3     5     2     4     1     3     5     2
    >>> return_paqs(df, incl_ids=False, other_cols=['OtherCol'])
       PAQ1  PAQ2  PAQ3  PAQ4  PAQ5  PAQ6  PAQ7  PAQ8 OtherCol
    0     4     2     1     3     5     2     4     1        A
    1     3     5     2     4     1     3     5     2        B
    """
    cols = PAQ_IDS.copy()

    if incl_ids:
        id_cols = [
            name
            for name in ["RecordID", "GroupID", "SessionID", "LocationID"]
            if name in df.columns
        ]
        cols = id_cols + cols

    if other_cols:
        cols.extend(other_cols)

    logger.debug(f"Returning PAQ columns: {cols}")
    return df[cols]


def rename_paqs(
    df: pd.DataFrame, paq_aliases: Union[Tuple, Dict] = None
) -> pd.DataFrame:
    """
    Rename the PAQ columns in a DataFrame to standard PAQ IDs.

    Parameters
    ----------
    df : pd.DataFrame
        Input DataFrame containing PAQ data.
    paq_aliases : Union[Tuple, Dict], optional
        Specify which PAQs are to be renamed. If None, will check if the column names
        are in pre-defined options. If a tuple, the order must match PAQ_IDS.
        If a dict, keys are current names and values are desired PAQ IDs.

    Returns
    -------
    pd.DataFrame
        DataFrame with renamed PAQ columns.

    Raises
    ------
    ValueError
        If paq_aliases is not a tuple, list, or dictionary.

    Examples
    --------
    >>> import pandas as pd
    >>> df = pd.DataFrame({
    ...     'pleasant': [4, 3],
    ...     'vibrant': [2, 5],
    ...     'other_col': [1, 2]
    ... })
    >>> rename_paqs(df)
       PAQ1  PAQ2  other_col
    0     4     2          1
    1     3     5          2
    >>> df_custom = pd.DataFrame({
    ...     'pl': [4, 3],
    ...     'vb': [2, 5],
    ... })
    >>> rename_paqs(df_custom, paq_aliases={'pl': 'PAQ1', 'vb': 'PAQ2'})
       PAQ1  PAQ2
    0     4     2
    1     3     5
    """
    if paq_aliases is None:
        if any(paq_id in df.columns for paq_id in PAQ_IDS):
            logger.info("PAQs already correctly named.")
            return df
        if any(paq_name in df.columns for paq_name in PAQ_LABELS):
            paq_aliases = PAQ_LABELS

    if isinstance(paq_aliases, (list, tuple)):
        rename_dict = dict(zip(paq_aliases, PAQ_IDS))
    elif isinstance(paq_aliases, dict):
        rename_dict = paq_aliases
    else:
        raise ValueError("paq_aliases must be a tuple, list, or dictionary.")

    logger.debug(f"Renaming PAQs with the following mapping: {rename_dict}")
    return df.rename(columns=rename_dict)


def mean_responses(df: pd.DataFrame, group: str) -> pd.DataFrame:
    """
    Calculate the mean responses for each PAQ group.

    Parameters
    ----------
    df : pd.DataFrame
        Input DataFrame containing PAQ data.
    group : str
        Column name to group by.

    Returns
    -------
    pd.DataFrame
        DataFrame with mean responses for each PAQ group.

    """
    df = return_paqs(df, incl_ids=False, other_cols=[group])
    return df.groupby(group).mean().reset_index()


# Add other utility functions here as needed



import pytest
import yaml
from pydantic import ValidationError

from soundscapy.audio.analysis_settings import (
    AnalysisSettings,
    ConfigManager,
    LibrarySettings,
    MetricSettings,
)


@pytest.fixture
def sample_config():
    return {
        "version": "1.0",
        "PythonAcoustics": {
            "LAeq": {
                "run": True,
                "main": "avg",
                "statistics": [5, 10, 50, 90, 95, "min", "max", "kurt", "skew"],
                "channel": ["Left", "Right"],
                "label": "LAeq",
                "func_args": {"time": 0.125, "method": "average"},
            }
        },
        "MoSQITo": {
            "loudness_zwtv": {
                "run": True,
                "main": 5,
                "statistics": [10, 50, 90, 95, "min", "max", "kurt", "skew", "avg"],
                "channel": ["Left", "Right"],
                "label": "N",
                "parallel": True,
                "func_args": {"field_type": "free"},
            }
        },
        "scikit-maad": {
            "all_temporal_alpha_indices": {"run": True, "channel": ["Left", "Right"]}
        },
    }


@pytest.fixture
def temp_config_file(tmp_path, sample_config):
    config_file = tmp_path / "test_config.yaml"
    with open(config_file, "w") as f:
        yaml.dump(sample_config, f)
    return config_file


class TestMetricSettings:
    def test_valid_metric_settings(self):
        settings = MetricSettings(
            run=True,
            main="avg",
            statistics=[5, 10, 50, 90, 95, "min", "max", "kurt", "skew"],
            channel=["Left", "Right"],
            label="LAeq",
            func_args={"time": 0.125, "method": "average"},
        )
        assert settings.run is True
        assert settings.main == "avg"
        assert "Left" in settings.channel and "Right" in settings.channel

    def test_invalid_metric_settings(self):
        with pytest.raises(ValidationError):
            MetricSettings(run="not_a_boolean")


class TestLibrarySettings:
    def test_valid_library_settings(self):
        settings = LibrarySettings(
            root={"LAeq": MetricSettings(run=True, main="avg", label="LAeq")}
        )
        assert "LAeq" in settings.root
        assert settings.root["LAeq"].run is True

    def test_invalid_library_settings(self):
        with pytest.raises(AttributeError):
            LibrarySettings(root={"InvalidMetric": "Not a MetricSettings object"})

    def test_get_existing_metric_settings(self, sample_config):
        settings = AnalysisSettings(**sample_config)
        library_settings = settings.PythonAcoustics
        result = library_settings.get_metric_settings("LAeq")
        assert result == library_settings.root["LAeq"]

    def test_get_non_existing_metric_settings(self):
        library_settings = LibrarySettings(root={})
        with pytest.raises(KeyError):
            library_settings.get_metric_settings("metric2")


class TestAnalysisSettings:
    def test_from_yaml(self, temp_config_file):
        settings = AnalysisSettings.from_yaml(temp_config_file)
        assert settings.version == "1.0"
        assert "LAeq" in settings.PythonAcoustics.root
        assert "loudness_zwtv" in settings.MoSQITo.root
        assert "all_temporal_alpha_indices" in settings.scikit_maad.root

    def test_to_yaml(self, tmp_path, sample_config):
        settings = AnalysisSettings(**sample_config)
        output_file = tmp_path / "output_config.yaml"
        settings.to_yaml(output_file)
        assert output_file.exists()

        # Read back and verify
        with open(output_file, "r") as f:
            loaded_config = yaml.safe_load(f)
        assert loaded_config["version"] == "1.0"
        assert "LAeq" in loaded_config["PythonAcoustics"]

    def test_get_metric_settings(self, sample_config):
        settings = AnalysisSettings(**sample_config)
        laeq_settings = settings.get_metric_settings("PythonAcoustics", "LAeq")
        assert isinstance(laeq_settings, MetricSettings)
        assert laeq_settings.run is True
        assert laeq_settings.main == "avg"

    def test_get_metric_settings_invalid(self, sample_config):
        settings = AnalysisSettings(**sample_config)
        with pytest.raises(AttributeError):
            settings.get_metric_settings("InvalidLibrary", "InvalidMetric")

    def test_get_enabled_metrics(self, sample_config):
        settings = AnalysisSettings(**sample_config)
        enabled = settings.get_enabled_metrics()
        assert "LAeq" in enabled["PythonAcoustics"]
        assert "loudness_zwtv" in enabled["MoSQITo"]
        assert "all_temporal_alpha_indices" in enabled["scikit_maad"]

    def test_update_existing_metric_setting(self, sample_config):
        settings = AnalysisSettings(**sample_config)
        settings.update_setting("PythonAcoustics", "LAeq", run=False)
        updated_metric = settings.get_metric_settings("PythonAcoustics", "LAeq")
        assert updated_metric.run is False

    def test_update_non_existing_metric_setting(self, sample_config):
        settings = AnalysisSettings(**sample_config)
        with pytest.raises(KeyError):
            settings.update_setting("PythonAcoustics", "metric2", run=False)

    def test_update_metric_with_invalid_setting(self, sample_config):
        settings = AnalysisSettings(**sample_config)
        with pytest.raises(KeyError):
            settings.update_setting("PythonAcoustics", "metric1", invalid_setting=True)


class TestConfigManager:
    @pytest.fixture
    def config_manager(self, temp_config_file):
        return ConfigManager(temp_config_file)

    def test_load_config(self, config_manager):
        config = config_manager.load_config()
        assert isinstance(config, AnalysisSettings)
        assert config.version == "1.0"

    def test_save_config(self, config_manager, tmp_path):
        config_manager.load_config()
        new_file = tmp_path / "new_config.yaml"
        config_manager.save_config(new_file)
        assert new_file.exists()

    def test_merge_configs(self, config_manager):
        config_manager.load_config()
        override = {"PythonAcoustics": {"LAeq": {"run": False}}}
        merged = config_manager.merge_configs(override)
        assert merged.PythonAcoustics.root["LAeq"].run is False

    def test_generate_minimal_config(self, config_manager):
        config_manager.load_config()
        minimal = config_manager.generate_minimal_config()
        assert "version" not in minimal  # Assuming version is the same as default
        assert "MoSQITo" in minimal

    def test_load_default_config(self):
        manager = ConfigManager()
        config = manager.load_config()
        assert isinstance(config, AnalysisSettings)
        # Add more assertions based on your default config structure


def test_end_to_end(temp_config_file, tmp_path):
    # Load configuration
    manager = ConfigManager(temp_config_file)
    config = manager.load_config()
    assert config.PythonAcoustics.root["LAeq"].run is True

    # Modify configuration
    override = {"PythonAcoustics": {"LAeq": {"run": False}}}
    merged_config = manager.merge_configs(override)
    assert merged_config.PythonAcoustics.root["LAeq"].run is False
    assert manager.current_config.PythonAcoustics.root["LAeq"].run is False

    # Save modified configuration
    new_file = tmp_path / "modified_config.yaml"
    manager.save_config(new_file)

    # Load the saved configuration and verify changes
    new_manager = ConfigManager(new_file)
    new_config = new_manager.load_config()
    assert new_config.PythonAcoustics.root["LAeq"].run is False


if __name__ == "__main__":
    pytest.main()



# Writing a basic test suite for the 'metrics.py' file using the unittest framework

import unittest

import numpy as np
import pytest

from soundscapy.audio.metrics import _stat_calcs


class TestMetricsUnit(unittest.TestCase):
    def test_stat_calcs_avg(self):
        ts = np.array([1, 2, 3, 4, 5])
        res = {}
        updated_res = _stat_calcs("metric", ts, res, ["avg"])
        self.assertEqual(
            updated_res["metric_avg"], 3.0, "Average calculation is incorrect"
        )

    def test_stat_calcs_percentile(self):
        ts = np.array([1, 2, 3, 4, 5])
        res = {}
        updated_res = _stat_calcs("metric", ts, res, [50])
        self.assertEqual(
            updated_res["metric_50"], 3.0, "50th percentile calculation is incorrect"
        )

    def test_stat_calcs_max(self):
        ts = np.array([1, 2, 3, 4, 5])
        res = {}
        updated_res = _stat_calcs("metric", ts, res, ["max"])
        self.assertEqual(
            updated_res["metric_max"], 5, "Max value calculation is incorrect"
        )

    def test_stat_calcs_multiple_stats(self):
        ts = np.array([1, 2, 3, 4, 5])
        res = {}
        updated_res = _stat_calcs("metric", ts, res, [50, "avg", "max"])
        expected_res = {"metric_50": 3.0, "metric_avg": 3.0, "metric_max": 5}
        self.assertEqual(
            updated_res, expected_res, "Multiple statistics calculation is incorrect"
        )

    @pytest.mark.filterwarnings("ignore:Mean of empty slice")
    @pytest.mark.filterwarnings("ignore:invalid value encountered in scalar divide")
    def test_stat_calcs_empty_ts(self):
        ts = np.array([])
        res = {}
        updated_res = _stat_calcs("metric", ts, res, ["avg"])
        self.assertTrue(
            np.isnan(updated_res.get("metric_avg")),
            "Average for empty array should be np.nan",
        )



from pathlib import Path

import pandas as pd
import pytest
import yaml

from soundscapy.audio.analysis_settings import AnalysisSettings
from soundscapy.audio.audio_analysis import AudioAnalysis

TEST_AUDIO_DIR = Path(__file__).parent.parent / "test_audio_files"


@pytest.fixture
def sample_config():
    return {
        "version": "1.0",
        "PythonAcoustics": {
            "LAeq": {
                "run": True,
                "main": "avg",
                "statistics": [5, 10, 50, 90, 95, "avg", "min", "max", "kurt", "skew"],
                "channel": ["Left", "Right"],
                "label": "LAeq",
                "func_args": {"time": 0.125, "method": "average"},
            }
        },
        "scikit-maad": {
            "all_temporal_alpha_indices": {"run": True, "channel": ["Left", "Right"]}
        },
    }


@pytest.fixture
def temp_config_file(tmp_path, sample_config):
    config_file = tmp_path / "test_config.yaml"
    with open(config_file, "w") as f:
        yaml.dump(sample_config, f)
    return config_file


@pytest.fixture
def audio_analysis(temp_config_file):
    return AudioAnalysis(temp_config_file)


class TestAudioAnalysis:
    def test_initialization(self, audio_analysis, temp_config_file):
        assert isinstance(
            audio_analysis.config_manager.current_config, AnalysisSettings
        )
        assert audio_analysis.config_manager.config_path == temp_config_file

    def test_analyze_file(self, audio_analysis):
        test_file = TEST_AUDIO_DIR / "trimmed_CT101.wav"
        result = audio_analysis.analyze_file(test_file)
        assert isinstance(result, pd.DataFrame)
        assert "LAeq" in result.columns
        assert "LEQt" in result.columns  # One of the MAAD indices

    @pytest.mark.slow
    def test_analyze_folder_single_worker(self, audio_analysis):
        # Count the number of WAV files in the test directory
        expected_file_count = len(list(TEST_AUDIO_DIR.glob("*.wav")))

        result = audio_analysis.analyze_folder(TEST_AUDIO_DIR, max_workers=1)
        assert isinstance(result, pd.DataFrame)
        assert len(result) > 0, "No results were produced"
        assert "LAeq" in result.columns
        assert "LEQt" in result.columns

        # Check if the number of rows in the result matches the number of WAV files
        analyzed_files = set(result.index.get_level_values(0))
        assert (
            len(analyzed_files) == expected_file_count
        ), f"Expected {expected_file_count} files to be analyzed, but got {len(analyzed_files)}"

        # Check if all expected files were analyzed
        expected_files = {f.stem for f in TEST_AUDIO_DIR.glob("*.wav")}
        assert (
            analyzed_files == expected_files
        ), f"Mismatch in analyzed files. Expected: {expected_files}, Got: {analyzed_files}"

    @pytest.mark.slow
    def test_analyze_folder_multi_worker(self, audio_analysis):
        # Count the number of WAV files in the test directory
        expected_file_count = len(list(TEST_AUDIO_DIR.glob("*.wav")))

        result = audio_analysis.analyze_folder(TEST_AUDIO_DIR, max_workers=None)
        assert isinstance(result, pd.DataFrame)
        assert len(result) > 0, "No results were produced"
        assert "LAeq" in result.columns
        assert "LEQt" in result.columns

        # Check if the number of rows in the result matches the number of WAV files
        analyzed_files = set(result.index.get_level_values(0))
        assert (
            len(analyzed_files) == expected_file_count
        ), f"Expected {expected_file_count} files to be analyzed, but got {len(analyzed_files)}"

        # Check if all expected files were analyzed
        expected_files = {f.stem for f in TEST_AUDIO_DIR.glob("*.wav")}
        assert (
            analyzed_files == expected_files
        ), f"Mismatch in analyzed files. Expected: {expected_files}, Got: {analyzed_files}"

    def test_save_results(self, audio_analysis, tmp_path):
        df = pd.DataFrame({"test": [1, 2]})
        csv_path = tmp_path / "results.csv"
        audio_analysis.save_results(df, csv_path)
        assert csv_path.exists()

        xlsx_path = tmp_path / "results.xlsx"
        audio_analysis.save_results(df, xlsx_path)
        assert xlsx_path.exists()

        with pytest.raises(ValueError):
            audio_analysis.save_results(df, tmp_path / "results.txt")

    def test_update_config(self, audio_analysis):
        new_config = {"PythonAcoustics": {"LAeq": {"run": False}}}
        audio_analysis.update_config(new_config)
        assert not audio_analysis.settings.PythonAcoustics.root["LAeq"].run

    def test_save_config(self, audio_analysis, tmp_path):
        config_path = tmp_path / "new_config.yaml"
        audio_analysis.save_config(config_path)
        assert config_path.exists()

    @pytest.mark.slow
    def test_mosqito_metric(self, audio_analysis):
        mosqito_config = {
            "MoSQITo": {
                "loudness_zwtv": {
                    "run": True,
                    "main": 5,
                    "statistics": [
                        5,
                        10,
                        50,
                        90,
                        95,
                        "min",
                        "max",
                        "kurt",
                        "skew",
                        "avg",
                    ],
                    "channel": ["Left", "Right"],
                    "label": "N",
                    "parallel": True,
                    "func_args": {"field_type": "free"},
                }
            }
        }
        audio_analysis.update_config(mosqito_config)

        test_file = TEST_AUDIO_DIR / "trimmed_CT101.wav"
        result = audio_analysis.analyze_file(test_file, resample=48000)

        assert isinstance(result, pd.DataFrame)
        assert "N_5" in result.columns


if __name__ == "__main__":
    pytest.main()



"""
Styling utilities for circumplex plots using Seaborn and Matplotlib.
"""

from dataclasses import dataclass, field
from typing import Any, Dict, Tuple

import matplotlib as mpl
import seaborn as sns

from .plotting_utils import DEFAULT_FIGSIZE


@dataclass
class StyleOptions:
    """
    Configuration options for styling circumplex plots.

    Attributes:
        diag_lines_zorder (int): Z-order for diagonal lines.
        diag_labels_zorder (int): Z-order for diagonal labels.
        prim_lines_zorder (int): Z-order for primary lines.
        data_zorder (int): Z-order for plotted data.
        bw_adjust (float): Bandwidth adjustment for kernel density estimation.
        figsize (Tuple[int, int]): Figure size (width, height) in inches.
        simple_density (Dict[str, Any]): Configuration for simple density plots.
    """

    diag_lines_zorder: int = 1
    diag_labels_zorder: int = 4
    prim_lines_zorder: int = 2
    data_zorder: int = 3
    bw_adjust: float = 1.2
    figsize: Tuple[int, int] = DEFAULT_FIGSIZE
    simple_density: Dict[str, Any] = field(
        default_factory=lambda: {
            "thresh": 0.5,
            "levels": 2,
            "incl_outline": True,
            "alpha": 0.5,
        }
    )


class SeabornStyler:
    """
    Class for applying Seaborn styles to circumplex plots.
    """

    def __init__(self, params: Any, style_options: StyleOptions = StyleOptions()):
        self.params = params
        self.style_options = style_options

    def apply_styling(
        self, fig: mpl.figure.Figure, ax: mpl.axes.Axes
    ) -> Tuple[mpl.figure.Figure, mpl.axes.Axes]:
        """
        Apply styling to the plot.

        Parameters
        ----------
            fig (mpl.figure.Figure): The figure object.
            ax (mpl.axes.Axes): The axes object.

        Returns
        -------
            Tuple[mpl.figure.Figure, mpl.axes.Axes]: The styled figure and axes.
        """
        self.set_style()
        self.circumplex_grid(ax)
        self.set_circum_title(ax)
        self.deal_w_default_labels(ax)
        if self.params.hue and self.params.legend is not False:
            self.move_legend(ax)
        return fig, ax

    def set_style(self) -> None:
        """Set the overall style for the plot."""
        sns.set_style({"xtick.direction": "in", "ytick.direction": "in"})

    def circumplex_grid(self, ax: mpl.axes.Axes) -> None:
        """Add the circumplex grid to the plot."""
        ax.set_xlim(self.params.xlim)
        ax.set_ylim(self.params.ylim)

        ax.get_xaxis().set_minor_locator(mpl.ticker.AutoMinorLocator())
        ax.get_yaxis().set_minor_locator(mpl.ticker.AutoMinorLocator())

        ax.grid(visible=True, which="major", color="grey", alpha=0.5)
        ax.grid(
            visible=True,
            which="minor",
            color="grey",
            linestyle="dashed",
            linewidth=0.5,
            alpha=0.4,
            zorder=self.style_options.prim_lines_zorder,
        )

        self.primary_lines_and_labels(ax)
        if self.params.diagonal_lines:
            self.diagonal_lines_and_labels(ax)

    def set_circum_title(self, ax: mpl.axes.Axes) -> None:
        """Set the title for the circumplex plot."""
        title_pad = 6.0
        ax.set_title(self.params.title, pad=title_pad)

    def deal_w_default_labels(self, ax: mpl.axes.Axes) -> None:
        """Handle the default labels for the axes."""
        if not self.params.show_labels:
            ax.set_xlabel("")
            ax.set_ylabel("")
        else:
            ax.set_xlabel(self.params.x)
            ax.set_ylabel(self.params.y)

    def move_legend(self, ax: mpl.axes.Axes) -> None:
        """Move the legend to the specified location."""
        old_legend = ax.get_legend()
        handles = old_legend.legend_handles
        labels = [t.get_text() for t in old_legend.get_texts()]
        title = old_legend.get_title().get_text()
        ax.legend(handles, labels, loc=self.params.legend_location, title=title)

    def primary_lines_and_labels(self, ax: mpl.axes.Axes) -> None:
        """Add primary lines to the plot."""
        line_weights = 1.5

        ax.axhline(
            y=0,
            color="grey",
            linestyle="dashed",
            alpha=1,
            lw=line_weights,
            zorder=self.style_options.prim_lines_zorder,
        )
        ax.axvline(
            x=0,
            color="grey",
            linestyle="dashed",
            alpha=1,
            lw=line_weights,
            zorder=self.style_options.prim_lines_zorder,
        )

    def diagonal_lines_and_labels(self, ax: mpl.axes.Axes) -> None:
        """Add diagonal lines and labels to the plot."""
        x_lim, y_lim = ax.get_xlim(), ax.get_ylim()
        line_weights = 1.5

        ax.plot(
            [x_lim[0], x_lim[1]],
            [y_lim[0], y_lim[1]],
            linestyle="dashed",
            color="grey",
            alpha=0.5,
            lw=line_weights,
            zorder=self.style_options.diag_lines_zorder,
        )
        ax.plot(
            [x_lim[0], x_lim[1]],
            [y_lim[1], y_lim[0]],
            linestyle="dashed",
            color="grey",
            alpha=0.5,
            lw=line_weights,
            zorder=self.style_options.diag_lines_zorder,
        )

        diag_ax_font = {
            "fontstyle": "italic",
            "fontsize": "small",
            "fontweight": "bold",
            "color": "black",
            "alpha": 0.5,
        }
        ax.text(
            x_lim[1] / 2,
            y_lim[1] / 2,
            "(vibrant)",
            ha="center",
            va="center",
            fontdict=diag_ax_font,
            zorder=self.style_options.diag_labels_zorder,
        )
        ax.text(
            x_lim[0] / 2,
            y_lim[1] / 2,
            "(chaotic)",
            ha="center",
            va="center",
            fontdict=diag_ax_font,
            zorder=self.style_options.diag_labels_zorder,
        )
        ax.text(
            x_lim[0] / 2,
            y_lim[0] / 2,
            "(monotonous)",
            ha="center",
            va="center",
            fontdict=diag_ax_font,
            zorder=self.style_options.diag_labels_zorder,
        )
        ax.text(
            x_lim[1] / 2,
            y_lim[0] / 2,
            "(calm)",
            ha="center",
            va="center",
            fontdict=diag_ax_font,
            zorder=self.style_options.diag_labels_zorder,
        )



import warnings
from abc import ABC, abstractmethod

import matplotlib.pyplot as plt
import plotly.express as px
import plotly.graph_objects as go
import seaborn as sns

from soundscapy.plotting.stylers import SeabornStyler, StyleOptions


class PlotBackend(ABC):
    """
    Abstract base class for plot backends.

    This class defines the interface for creating scatter and density plots,
    as well as applying styling to the plots.
    """

    @abstractmethod
    def create_scatter(self, data, params):
        """
        Create a scatter plot.

        Parameters
        ----------
            data (pd.DataFrame): The data to plot.
            params (CircumplexPlotParams): The parameters for the plot.

        Returns
        -------
            The created plot object.
        """
        pass

    @abstractmethod
    def create_density(self, data, params):
        """
        Create a density plot.

        Parameters
        ----------
            data (pd.DataFrame): The data to plot.
            params (CircumplexPlotParams): The parameters for the plot.

        Returns
        -------
            The created plot object.
        """
        pass

    @abstractmethod
    def apply_styling(self, plot_obj, params):
        """
        Apply styling to the plot.

        Parameters
        ----------
            plot_obj: The plot object to style.
            params (CircumplexPlotParams): The parameters for styling.

        Returns
        -------
            The styled plot object.
        """
        pass


class SeabornBackend(PlotBackend):
    """
    Backend for creating plots using Seaborn and Matplotlib.
    """

    def __init__(self, style_options: StyleOptions = StyleOptions()):
        self.style_options = style_options

    def create_scatter(self, data, params, ax=None):
        """
        Create a scatter plot using Seaborn.

        Parameters
        ----------
            data (pd.DataFrame): The data to plot.
            params (CircumplexPlotParams): The parameters for the plot.

        Returns
        -------
            tuple: A tuple containing the figure and axes objects.
        """
        if ax is None:
            fig, ax = plt.subplots(figsize=self.style_options.figsize)
        else:
            fig = ax.get_figure()
        if "incl_scatter" in params.extra_params:
            params.extra_params.pop("incl_scatter")
        if "density_type" in params.extra_params:
            params.extra_params.pop("density_type")
        if "fill" in params.extra_params:
            params.extra_params.pop("fill")
        sns.scatterplot(
            data=data,
            x=params.x,
            y=params.y,
            hue=params.hue,
            palette=params.palette if params.hue else None,
            alpha=params.alpha,
            ax=ax,
            zorder=self.style_options.data_zorder,
            legend=params.legend,
            **params.extra_params,
        )
        return fig, ax

    def create_density(self, data, params, ax=None):
        """
        Create a density plot using Seaborn.

        Parameters
        ----------
            data (pd.DataFrame): The data to plot.
            params (CircumplexPlotParams): The parameters for the plot.

        Returns
        -------
            tuple: A tuple containing the figure and axes objects.
        """
        if len(data) < 30:
            warnings.warn(
                "Density plots are not recommended for small datasets (<30 samples).",
                UserWarning,
            )

        if ax is None:
            fig, ax = plt.subplots(figsize=self.style_options.figsize)
        else:
            fig = ax.get_figure()
        sns.kdeplot(
            data=data,
            x=params.x,
            y=params.y,
            hue=params.hue,
            palette=params.palette,
            fill=params.fill,
            alpha=params.alpha,
            ax=ax,
            bw_adjust=self.style_options.bw_adjust,
            zorder=self.style_options.data_zorder,
            common_norm=False,
            legend=params.legend,
            **params.extra_params,
        )
        if params.incl_outline:
            sns.kdeplot(
                data=data,
                x=params.x,
                y=params.y,
                hue=params.hue,
                alpha=1,
                palette=params.palette,
                fill=False,
                ax=ax,
                bw_adjust=self.style_options.bw_adjust,
                zorder=self.style_options.data_zorder,
                common_norm=False,
                legend=False,
                **params.extra_params,
            )
        return fig, ax

    def create_jointplot(self, data, params):
        """

        Examples
        --------
        >>> import soundscapy as sspy
        >>> from soundscapy.plotting import Backend, CircumplexPlot, StyleOptions, CircumplexPlotParams
        >>> data = sspy.isd.load()
        >>> data = sspy.surveys.add_iso_coords(data, overwrite=True)
        >>> sample_data = sspy.isd.select_location_ids(data, ['CamdenTown'])
        >>> plot = CircumplexPlot(data=sample_data, backend=Backend.SEABORN)
        >>> g = plot.jointplot()
        >>> g.show() # doctest: +SKIP
        """
        g = sns.JointGrid(xlim=params.xlim, ylim=params.ylim)
        joint_params = params
        joint_params.title = ""
        SeabornBackend.create_density(self, data, joint_params, ax=g.ax_joint)

        margin_params = params
        margin_params.title = ""
        sns.kdeplot(data, x=params.x, ax=g.ax_marg_x, fill=True, alpha=params.alpha)
        sns.kdeplot(data, y=params.y, ax=g.ax_marg_y, fill=True, alpha=params.alpha)

        return (
            g.fig,
            g.ax_joint,
        )  # TODO: Should return the whole JointGrid object - repeat throughout plotting methods

    def create_simple_density(self, data, params, ax=None):
        if ax is None:
            fig, ax = plt.subplots(figsize=self.style_options.figsize)
        else:
            fig = ax.get_figure()

        sns.kdeplot(
            data=data,
            x=params.x,
            y=params.y,
            hue=params.hue,
            palette=params.palette,
            fill=params.fill,
            ax=ax,
            thresh=self.style_options.simple_density["thresh"],
            levels=self.style_options.simple_density["levels"],
            alpha=self.style_options.simple_density["alpha"],
            zorder=self.style_options.data_zorder,
            common_norm=False,
            legend=params.legend,
            **params.extra_params,
        )
        if params.incl_outline:
            sns.kdeplot(
                data=data,
                x=params.x,
                y=params.y,
                hue=params.hue,
                palette=params.palette,
                fill=False,
                ax=ax,
                thresh=self.style_options.simple_density["thresh"],
                levels=self.style_options.simple_density["levels"],
                zorder=self.style_options.data_zorder,
                alpha=1,
                common_norm=False,
                legend=False,
                **params.extra_params,
            )
        return fig, ax

    def apply_styling(self, plot_obj, params):
        """
        Apply styling to the Seaborn plot.

        Parameters
        ----------
            plot_obj (tuple): A tuple containing the figure and axes objects.
            params (CircumplexPlotParams): The parameters for styling.

        Returns
        -------
            tuple: The styled figure and axes objects.
        """
        fig, ax = plot_obj
        styler = SeabornStyler(params, self.style_options)
        return styler.apply_styling(fig, ax)

    def show(self, plot_obj):
        """
        Display the Matplotlib figure.

        Parameters
        ----------
            fig: The figure to display.
        """
        fig, _ = plot_obj
        plt.show()


class PlotlyBackend(PlotBackend):
    """
    Backend for creating plots using Plotly.
    """

    def __init__(self):
        warnings.warn(
            "PlotlyBackend is very experimental and not fully implemented.", UserWarning
        )
        pass

    def create_scatter(self, data, params):
        """
        Create a scatter plot using Plotly.

        Parameters
        ----------
            data (pd.DataFrame): The data to plot.
            params (CircumplexPlotParams): The parameters for the plot.

        Returns
        -------
            go.Figure: A Plotly figure object.
        """
        fig = px.scatter(
            data,
            x=params.x,
            y=params.y,
            color=params.hue,
            title=params.title,
            range_x=params.xlim,
            range_y=params.ylim,
            **params.extra_params,
        )
        fig.update_layout(
            width=600,
            height=600,
            xaxis=dict(scaleanchor="y", scaleratio=1),
            yaxis=dict(scaleanchor="x", scaleratio=1),
        )
        return fig

    def create_density(self, data, params):
        """
        Create a density plot using Plotly.

        Parameters
        ----------
            data (pd.DataFrame): The data to plot.
            params (CircumplexPlotParams): The parameters for the plot.

        Returns
        -------
            go.Figure: A Plotly figure object.
        """
        if len(data) < 30:
            warnings.warn(
                "Density plots are not recommended for small datasets (<30 samples). Consider using a scatter plot instead.",
                UserWarning,
            )

        fig = px.density_heatmap(
            data,
            x=params.x,
            y=params.y,
            title=params.title,
            range_x=params.xlim,
            range_y=params.ylim,
            **params.extra_params,
        )
        fig.update_layout(
            width=600,
            height=600,
            xaxis=dict(scaleanchor="y", scaleratio=1),
            yaxis=dict(scaleanchor="x", scaleratio=1),
        )
        scatter_trace = px.scatter(
            data, x=params.x, y=params.y, color=params.hue, opacity=0.5
        ).data[0]
        fig.add_trace(scatter_trace)
        return fig

    def apply_styling(self, plot_obj, params):
        """
        Apply styling to the Plotly plot.

        Parameters
        ----------
            plot_obj (go.Figure): A Plotly figure object.
            params (CircumplexPlotParams): The parameters for styling.

        Returns
        -------
            go.Figure: The styled Plotly figure object.
        """
        fig = plot_obj
        if params.diagonal_lines:
            fig.add_trace(
                go.Scatter(
                    x=[params.xlim[0], params.xlim[1]],
                    y=[params.ylim[0], params.ylim[1]],
                    mode="lines",
                    line=dict(color="gray", dash="dash"),
                    showlegend=False,
                )
            )
            fig.add_trace(
                go.Scatter(
                    x=[params.xlim[0], params.xlim[1]],
                    y=[params.ylim[1], params.ylim[0]],
                    mode="lines",
                    line=dict(color="gray", dash="dash"),
                    showlegend=False,
                )
            )
        return fig

    def show(self, fig):
        """
        Display the Plotly figure.

        Parameters
        ----------
            fig (go.Figure): The Plotly figure to display.
        """
        fig.show()



"""
Soundscapy is a Python library for soundscape analysis and visualisation.
"""

# ruff: noqa: E402
from typing import Any
from loguru import logger

# https://loguru.readthedocs.io/en/latest/resources/recipes.html#configuring-loguru-to-be-used-by-a-library-or-an-application
logger.disable("soundscapy")

import importlib.metadata

__version__ = importlib.metadata.version("soundscapy")

from soundscapy._optionals import import_optional

# Always available core modules
from soundscapy import surveys
from soundscapy import databases
from soundscapy import plotting
from soundscapy.logging import setup_logging
from soundscapy.databases import araus, isd, satp
from soundscapy.surveys import processing
from soundscapy.plotting import scatter_plot, density_plot

__all__ = [
    # Core modules
    "surveys",
    "databases",
    "plotting",
    "araus",
    "isd",
    "satp",
    "processing",
    "scatter_plot",
    "density_plot",
    "setup_logging",
    # Optional modules listed explicitly for IDE/typing support
    "Binaural",
    "AudioAnalysis",
    "AnalysisSettings",
    "ConfigManager",
    "process_all_metrics",
    "prep_multiindex_df",
    "add_results",
    "parallel_process",
]


def __getattr__(name: str) -> Any:
    """Lazy import handling for optional components."""
    return import_optional(name)



"""
soundscapy.audio.parallel_processing
====================================

This module provides functions for parallel processing of binaural audio files.

It includes functions to load and analyze binaural files, as well as to process
multiple files in parallel using concurrent.futures.

Functions:
    load_analyse_binaural: Load and analyze a single binaural file.
    parallel_process: Process multiple binaural files in parallel.

Note:
    This module requires the tqdm library for progress bars and concurrent.futures
    for parallel processing. It uses loguru for logging.
"""

import concurrent.futures
from pathlib import Path
from typing import Dict, List, Optional
from loguru import logger

import pandas as pd
from tqdm.auto import tqdm

from .analysis_settings import AnalysisSettings
from .binaural import Binaural
from .metrics import (
    add_results,
    prep_multiindex_df,
    process_all_metrics,
)


def tqdm_write_sink(message):
    """
    A custom sink for loguru that writes messages using tqdm.write().

    This ensures that log messages don't interfere with tqdm progress bars.
    """
    tqdm.write(message, end="")


def load_analyse_binaural(
    wav_file: Path,
    levels: Dict | List[float],
    analysis_settings: AnalysisSettings,
    parallel_mosqito: bool = True,
    resample: Optional[int] = None,
) -> pd.DataFrame:
    """
    Load and analyze a single binaural audio file.

    Parameters
    ----------
    resample
    wav_file : Path
        Path to the WAV file.
    levels : Dict
        Dictionary with calibration levels for each channel.
    analysis_settings : AnalysisSettings
        Analysis settings object.
    parallel_mosqito : bool, optional
        Whether to process MoSQITo metrics in parallel. Defaults to True.

    Returns
    -------
    pd.DataFrame
        DataFrame with analysis results.
    """
    logger.info(f"Processing {wav_file}")
    try:
        b = Binaural.from_wav(wav_file, resample=resample)
        if levels is not None:
            if isinstance(levels, dict) and b.recording in levels:
                decibel = (levels[b.recording]["Left"], levels[b.recording]["Right"])
                b.calibrate_to(decibel, inplace=True)
            elif isinstance(levels, list | tuple):
                logger.debug(f"Calibrating {wav_file} to {levels} dB")
                b.calibrate_to(levels, inplace=True)
            else:
                logger.warning(f"No calibration levels found for {wav_file}")
        else:
            logger.warning(f"No calibration levels found for {wav_file}")
        return process_all_metrics(b, analysis_settings, parallel=parallel_mosqito)
    except Exception as e:
        logger.error(f"Error processing {wav_file}: {str(e)}")
        raise


def parallel_process(
    wav_files: List[Path],
    results_df: pd.DataFrame,
    levels: Dict,
    analysis_settings: AnalysisSettings,
    max_workers: Optional[int] = None,
    parallel_mosqito: bool = True,
    resample: Optional[int] = None,
) -> pd.DataFrame:
    """
    Process multiple binaural files in parallel.

    Parameters
    ----------
    resample
    wav_files : List[Path]
        List of WAV files to process.
    results_df : pd.DataFrame
        Initial results DataFrame to update.
    levels : Dict
        Dictionary with calibration levels for each file.
    analysis_settings : AnalysisSettings
        Analysis settings object.
    max_workers : int, optional
        Maximum number of worker processes. If None, it will default to the number of processors on the machine.
    parallel_mosqito : bool, optional
        Whether to process MoSQITo metrics in parallel within each file. Defaults to True.

    Returns
    -------
    pd.DataFrame
        Updated results DataFrame with analysis results for all files.
    """
    logger.info(f"Starting parallel processing of {len(wav_files)} files")

    # Add a handler that uses tqdm.write for output
    tqdm_handler_id = logger.add(tqdm_write_sink, format="{message}")

    with concurrent.futures.ProcessPoolExecutor(max_workers=max_workers) as executor:
        futures = []
        for wav_file in wav_files:
            future = executor.submit(
                load_analyse_binaural,
                wav_file,
                levels,
                analysis_settings,
                parallel_mosqito,
                resample,
            )
            futures.append(future)

        with tqdm(total=len(futures), desc="Processing files") as pbar:
            for future in concurrent.futures.as_completed(futures):
                try:
                    result = future.result()
                    results_df = add_results(results_df, result)
                except Exception as e:
                    logger.error(f"Error processing file: {str(e)}")
                finally:
                    pbar.update(1)

    # Remove the tqdm-compatible handler
    logger.remove(tqdm_handler_id)

    logger.info("Parallel processing completed")
    return results_df


if __name__ == "__main__":
    # Example usage
    from datetime import datetime
    import json
    import warnings

    warnings.filterwarnings("ignore")

    from soundscapy.logging import setup_logging

    setup_logging("DEBUG")

    base_path = Path().absolute().parent.parent.parent
    wav_folder = base_path.joinpath("test", "data")
    levels_file = wav_folder.joinpath("Levels.json")

    with open(levels_file) as f:
        levels = json.load(f)

    analysis_settings = AnalysisSettings.default()

    df = prep_multiindex_df(levels, incl_metric=True)

    wav_files = list(wav_folder.glob("*.wav"))

    df = parallel_process(wav_files[:4], df, levels, analysis_settings)

    output_file = base_path.joinpath(
        "test", f"ParallelTest_{datetime.now().strftime('%Y%m%d_%H%M%S')}.xlsx"
    )
    df.to_excel(output_file)
    print(f"Results saved to {output_file}")
